'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var compilerSfc = require('@vue/compiler-sfc');
var createDebugger = _interopDefault(require('debug'));
var hash = _interopDefault(require('hash-sum'));
var path = require('path');
var qs = _interopDefault(require('querystring'));
var rollupPluginutils = require('rollup-pluginutils');
var sourcemapCodec = require('sourcemap-codec');

try {
    require.resolve('@vue/compiler-sfc');
}
catch (e) {
    throw new Error('vue-loader requires @vue/compiler-sfc to be present in the dependency ' +
        'tree.');
}
const debug = createDebugger('rollup-plugin-vue');
const defaultOptions = {
    include: /\.vue$/,
    exclude: [],
    target: 'browser',
    exposeFilename: false,
};
function PluginVue(userOptions = {}) {
    const options = {
        ...defaultOptions,
        ...userOptions,
    };
    const isServer = options.target === 'node';
    const isProduction = process.env.NODE_ENV === 'production' || process.env.BUILD === 'production';
    const rootContext = process.cwd();
    const filter = rollupPluginutils.createFilter(options.include, options.exclude);
    return {
        name: 'vue',
        resolveId(id) {
            const query = parseVuePartRequest(id);
            if (query.vue) {
                debug(`resolveId(${id})`);
                return id;
            }
            return undefined;
        },
        load(id) {
            const query = parseVuePartRequest(id);
            if (query.vue) {
                const descriptor = getDescriptor(query.filename);
                const block = query.type === 'template'
                    ? descriptor.template
                    : query.type === 'script'
                        ? descriptor.script
                        : query.type === 'style'
                            ? descriptor.styles[query.index]
                            : query.type === 'custom'
                                ? descriptor.customBlocks[query.index]
                                : null;
                if (block) {
                    const result = {
                        code: block.content,
                        map: normalizeSourceMap(block.map),
                    };
                    if (query.type === 'template') {
                        // generate source mapping for each character.
                        result.map.mappings = sourcemapCodec.encode(result.code.split(/\r?\n/).map((line, index) => {
                            const segments = [];
                            for (let i = 0; i < line.length; ++i) {
                                segments.push([i, 0, block.loc.start.line + index - 1, i]);
                            }
                            return segments;
                        }));
                    }
                    debug(`load(${id})`, '\n' +
                        result.code +
                        '\n//# sourceMappingURL=data:application/json;charset=utf-8;base64,' +
                        Buffer.from(JSON.stringify(result.map), 'utf-8').toString('base64'));
                    return result;
                }
            }
            return undefined;
        },
        async transform(code, id) {
            const query = parseVuePartRequest(id);
            if (query.vue) {
                const descriptor = getDescriptor(query.filename);
                if (query.type === 'template') {
                    debug(`transform(${id})`);
                    const block = descriptor.template;
                    const result = compilerSfc.compileTemplate({
                        filename: query.filename,
                        source: code,
                        preprocessLang: block.lang,
                        compiler: options.compiler,
                        compilerOptions: {
                            ...options.compilerOptions,
                            scopeId: `data-v-${query.id}`,
                        },
                        transformAssetUrls: options.transformAssetUrls,
                    });
                    if (result.errors.length) {
                        result.errors.forEach(error => this.error(typeof error === 'string'
                            ? { id: query.filename, message: error }
                            : createRollupError(query.filename, error)));
                        return null;
                    }
                    if (result.tips.length) {
                        result.tips.forEach(tip => this.warn({
                            id: query.filename,
                            message: tip,
                        }));
                    }
                    return {
                        code: result.code,
                        map: normalizeSourceMap(result.map),
                    };
                }
                else if (query.type === 'style' && query.scoped) {
                    debug(`transform(${id})`);
                    const block = descriptor.styles[query.index];
                    const result = await compilerSfc.compileStyleAsync({
                        filename: query.filename,
                        id: `data-v-${query.id}`,
                        source: block.content,
                        scoped: query.scoped,
                        trim: true,
                    });
                    if (result.errors.length) {
                        result.errors.forEach(error => this.error({
                            id: query.filename,
                            message: error.message,
                        }));
                        return null;
                    }
                    return {
                        code: result.code,
                        map: normalizeSourceMap(result.map),
                    };
                }
                return null;
            }
            else if (filter(id)) {
                debug(`transform(${id})`);
                const { descriptor, errors } = parseSFC(code, id, rootContext);
                if (errors.length) {
                    errors.forEach(error => this.error(createRollupError(id, error)));
                    return null;
                }
                // module id for scoped CSS & hot-reload
                const output = transformVueSFC(code, id, descriptor, { rootContext, isProduction, isServer }, options);
                debug('transient .vue file:', '\n' + output + '\n');
                return {
                    code: output,
                    map: {
                        mappings: '',
                    },
                };
            }
            else {
                return null;
            }
        },
    };
}
function parseVuePartRequest(id) {
    const [filename, query] = id.split('?', 2);
    if (!query)
        return { vue: false, filename };
    const raw = qs.parse(query);
    if ('vue' in raw) {
        return {
            ...raw,
            filename,
            vue: true,
            type: raw.type,
            index: Number(raw.index),
            scoped: 'scoped' in raw,
            module: raw.module,
        };
    }
    return { vue: false, filename };
}
const cache = new Map();
function getDescriptor(id) {
    if (cache.has(id)) {
        return cache.get(id);
    }
    throw new Error(`${id} is not parsed it yet`);
}
function parseSFC(code, id, sourceRoot) {
    const { descriptor, errors } = compilerSfc.parse(code, {
        sourceMap: true,
        filename: id,
        sourceRoot: sourceRoot,
        pad: 'line',
    });
    cache.set(id, descriptor);
    return { descriptor, errors };
}
function transformVueSFC(code, resourcePath, descriptor, { rootContext, isProduction, }, options) {
    const shortFilePath = path.relative(rootContext, resourcePath)
        .replace(/^(\.\.[\/\\])+/, '')
        .replace(/\\/g, '/');
    const id = hash(isProduction ? shortFilePath + '\n' + code : shortFilePath);
    // feature information
    const hasScoped = descriptor.styles.some(s => s.scoped);
    const templateImport = getTemplateCode(descriptor, resourcePath, id, hasScoped);
    const scriptImport = getScriptCode(descriptor, resourcePath);
    const stylesCode = getStyleCode(descriptor, resourcePath, id);
    const output = [
        scriptImport,
        templateImport,
        stylesCode,
        `script.render = render`,
    ];
    if (hasScoped) {
        output.push(`script.__scopeId = ${_(`data-v-${id}`)}`);
    }
    if (!isProduction) {
        output.push(`script.__file = ${_(shortFilePath)}`);
    }
    else if (options.exposeFilename) {
        output.push(`script.__file = ${_(path.basename(shortFilePath))}`);
    }
    output.push('export default script');
    return output.join('\n');
}
function getTemplateCode(descriptor, resourcePath, id, hasScoped) {
    let templateImport = `const render = () => {}`;
    let templateRequest;
    if (descriptor.template) {
        const src = descriptor.template.src || resourcePath;
        const idQuery = `&id=${id}`;
        const scopedQuery = hasScoped ? `&scoped=true` : ``;
        const attrsQuery = attrsToQuery(descriptor.template.attrs);
        const query = `?vue&type=template${idQuery}${scopedQuery}${attrsQuery}`;
        templateRequest = _(src + query);
        templateImport = `import { render } from ${templateRequest}`;
    }
    return templateImport;
}
function getScriptCode(descriptor, resourcePath) {
    let scriptImport = `const script = {}`;
    if (descriptor.script) {
        const src = descriptor.script.src || resourcePath;
        const attrsQuery = attrsToQuery(descriptor.script.attrs, 'js');
        const query = `?vue&type=script${attrsQuery}`;
        const scriptRequest = _(src + query);
        scriptImport =
            `import script from ${scriptRequest}\n` + `export * from ${scriptRequest}`; // support named exports
    }
    return scriptImport;
}
function getStyleCode(descriptor, resourcePath, id) {
    let stylesCode = ``;
    let hasCSSModules = false;
    if (descriptor.styles.length) {
        descriptor.styles.forEach((style, i) => {
            const src = style.src || resourcePath;
            const attrsQuery = attrsToQuery(style.attrs, 'css');
            // make sure to only pass id when necessary so that we don't inject
            // duplicate tags when multiple components import the same css file
            const idQuery = style.scoped ? `&id=${id}` : ``;
            const query = `?vue&type=style&index=${i}${idQuery}${attrsQuery}`;
            const styleRequest = _(src + query);
            if (style.module) {
                if (!hasCSSModules) {
                    stylesCode += `const cssModules = script.__cssModules = {}`;
                    hasCSSModules = true;
                }
                stylesCode += genCSSModulesCode(id, i, styleRequest, style.module);
            }
            else {
                stylesCode += `\nimport ${styleRequest}`;
            }
            // TODO SSR critical CSS collection
        });
    }
    return stylesCode;
}
function createRollupError(id, error) {
    return {
        id,
        plugin: 'vue',
        pluginCode: String(error.code),
        message: error.message,
        frame: error.loc.source,
        parserError: error,
        loc: error.loc
            ? {
                file: id,
                line: error.loc.start.line,
                column: error.loc.start.column,
            }
            : undefined,
    };
}
// these are built-in query parameters so should be ignored
// if the user happen to add them as attrs
const ignoreList = ['id', 'index', 'src', 'type'];
function attrsToQuery(attrs, langFallback) {
    let query = ``;
    for (const name in attrs) {
        const value = attrs[name];
        if (!ignoreList.includes(name)) {
            query += `&${qs.escape(name)}=${value ? qs.escape(String(value)) : ``}`;
        }
    }
    if (langFallback && !(`lang` in attrs)) {
        query += `&lang.${langFallback}`;
    }
    return query;
}
function _(any) {
    return JSON.stringify(any);
}
function normalizeSourceMap(map) {
    if (!map)
        return null;
    return {
        version: Number(map.version),
        file: map.file,
        names: map.names,
        sources: map.sources,
        sourceRoot: map.sourceRoot,
        sourcesContent: map.sourcesContent,
        mappings: typeof map.mappings === 'string' ? map.mappings : '',
    };
}
function genCSSModulesCode(
// @ts-ignore
id, index, request, moduleName) {
    const styleVar = `style${index}`;
    let code = `\nimport ${styleVar} from ${request}`;
    // inject variable
    const name = typeof moduleName === 'string' ? moduleName : '$style';
    code += `\ncssModules["${name}"] = ${styleVar}`;
    return code;
}

module.exports = PluginVue;
//# sourceMappingURL=index.js.map
